// -*-c++-*-

/*!
  \file audio_codec.cpp
  \brief audio message encoder/decoder Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA and ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "audio_codec.h"

#include "server_param.h"

#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>
#include <rcsc/types.h>

#include <iostream>
#include <algorithm>
#include <limits>

namespace rcsc {

const double AudioCodec::ERROR_VALUE = std::numeric_limits< double >::max();

const double AudioCodec::X_NORM_FACTOR = 57.5;
const double AudioCodec::Y_NORM_FACTOR = 39.0;
const double AudioCodec::SPEED_NORM_FACTOR = 3.0;

const double AudioCodec::COORD_STEP_L2 = 0.1;
const double AudioCodec::SPEED_STEP_L1 = 0.1;


//[-0-9a-zA-Z ().+*/?<>_]
const std::string AudioCodec::CHAR_SET =
"0123456789"
"abcdefghijklmnopqrstuvwxyz"
"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
" ().+-*/?<>_";

const int AudioCodec::CHAR_SIZE = static_cast< int >( CHAR_SET.length() );


/*-------------------------------------------------------------------*/
/*!

*/
AudioCodec::AudioCodec()
{
    // create int <-> char map

    for ( int i = 0; i < CHAR_SIZE; ++i )
    {
        M_char_to_int_map.insert( std::make_pair( CHAR_SET[i], i ) );
        M_int_to_char_map.push_back( CHAR_SET[i] );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
const
AudioCodec &
AudioCodec::i()
{
    static AudioCodec s_instance;

    return s_instance;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::encodeInt64ToStr( const boost::int64_t & ival,
                              const int len,
                              std::string & to ) const
{
    std::vector< int > remainder_values;
    remainder_values.reserve( len );

    boost::int64_t divided = ival;

    for ( int i = 0; i < len - 1; ++i )
    {
        remainder_values.push_back( divided % CHAR_SIZE );
        divided /= CHAR_SIZE;
    }

    if ( divided >= CHAR_SIZE )
    {
        std::cerr << __FILE__ << ':' << __LINE__
                  << " ***ERROR*** AudioCodec::encodeInt64ToStr."
                  << " Illegal value. "
                  << std::endl;
        return false;
    }

    remainder_values.push_back( divided );

    try
    {
        const std::vector< int >::reverse_iterator end = remainder_values.rend();
        for ( std::vector< int >::reverse_iterator it = remainder_values.rbegin();
              it != end;
              ++it )
        {
            to += M_int_to_char_map.at( *it );
        }
    }
    catch ( std::exception & e )
    {
        std::cerr << __FILE__ << ':' << __LINE__
                  << " ***ERROR*** AudioCodec::encodeInt64ToStr."
                  << " Exception caught: " << e.what()
                  << std::endl;
        return false;
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::decodeStrToInt64( const std::string & from,
                              boost::int64_t * to ) const
{
    if ( from.empty() )
    {
        return false;
    }

    boost::int64_t rval = 0;
    int digit_count = from.length() - 1;

    const std::string::const_iterator end = from.end();
    for ( std::string::const_iterator ch = from.begin();
          ch != end;
          ++ch, --digit_count )
    {
        CharToIntCont::const_iterator it = M_char_to_int_map.find( *ch );
        if ( it == M_char_to_int_map.end() )
        {
            std::cerr << __FILE__ << ": " << __LINE__
                      << " ***ERROR*** AudioCodec::decodeStrToInt64."
                      << " Unexpected communication message. ["
                      << from << "]"
                      << std::endl;
            return false;
        }

        rval
            += static_cast< boost::int64_t >( it->second )
            * static_cast< boost::int64_t >
            ( std::pow( static_cast< double >( CHAR_SIZE ),
                        static_cast< double >( digit_count ) ) );
    }

    if ( to )
    {
        *to = rval;
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
char
AudioCodec::encodePercentageToChar( const double & value ) const
{
    if ( value < -0.000001 || 1.000001 < value )
    {
        std::cerr << __FILE__ << ':' << __LINE__
                  << " ***ERROR*** illegal value range = "
                  << value << std::endl;
        return '\0';
    }

    int ival = static_cast< int >( rint( value * ( CHAR_SIZE - 1 ) ) * 10000.0 );
    ival /= 10000;
    if ( ival >= CHAR_SIZE )
    {
        std::cerr << __FILE__ << ':' << __LINE__
                  << " ***ERROR*** generated illegal index = "
                  << ival << " <- " << value << std::endl;
        return '\0';
    }

    try
    {
        return  intToCharMap().at( ival );
    }
    catch ( ... )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::encodePercentageToChar."
                  << " Exception caught! Failed to encode"
                  << std::endl;
        return '\0';
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
double
AudioCodec::decodeCharToPercentage( const char ch ) const
{
    CharToIntCont::const_iterator it = charToIntMap().find( ch );
    if ( it == charToIntMap().end() )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::decodeCharToPercentage."
                  << " Unsupported character [" << ch << "]"
                  << std::endl;
        return ERROR_VALUE;
    }

    return ( static_cast< double >( it->second )
             / static_cast< double >( CHAR_SIZE - 1) );
}

/*-------------------------------------------------------------------*/
/*!

*/
boost::int32_t
AudioCodec::posToBit18( const Vector2D & pos ) const
{
    boost::int32_t rval = 0;

    // pos.x value -> 9 bits (=[0,511])
    {
        double x = min_max( -52.0, pos.x, 52.0 );
        x += 52.0;
        x *= ( 511.0 / 104.0 ); // x /= (104.0/511.0);
        rval |= static_cast< boost::int32_t >( rint( x ) );
    }

    rval <<= 9; // 9 bits shift for next info

    // pos.y value -> 9 bits (=[0,511])
    {
        double y = min_max( -34.0, pos.y, 34.0 );
        y += 34.0;
        y *= ( 511.0 / 68.0 ); // y /= (68.0/511.0);
        rval |= static_cast< boost::int32_t >( rint( y ) );
    }

    return rval;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
AudioCodec::bit18ToPos( const boost::int32_t & val,
                        Vector2D * pos ) const
{
    // pos.x
    {
        boost::int32_t ix = ( val >> 9 ) & MASK_9;
        pos->x = ix * ( 104.0 / 511.0 ) - 52.0;
    }

    // pos.y
    {
        boost::int32_t iy = val & MASK_9;
        pos->y = iy * ( 68.0 / 511.0 ) - 34.0;
    }
}


/*-------------------------------------------------------------------*/
/*!

*/
boost::int32_t
AudioCodec::posToBit19( const Vector2D & pos ) const
{
    boost::int32_t rval = 0;

    // pos.x value -> 10 bits (=[0,1023])
    {
        double x = min_max( -52.5, pos.x, 52.5 );
        x += 52.5;
        x *= ( 1023.0 / 105.0 ); // x /= (105.0/1023.0);
        rval |= static_cast< boost::int32_t >( rint( x ) );
    }

    rval <<= 9; // 9 bits shift for next info

    // pos.y value -> 9 bits (=[0,511])
    {
        double y = min_max( -34.0, pos.y, 34.0 );
        y += 34.0;
        y *= ( 511.0 / 68.0 ); // y /= (68.0/511.0);
        rval |= static_cast< boost::int32_t >( rint( y ) );
    }

    return rval;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
AudioCodec::bit19ToPos( const boost::int32_t & val,
                        Vector2D * pos ) const
{
    // pos.x
    {
        boost::int32_t ix = ( val >> 9 ) & MASK_10;
        pos->x = ix * ( 105.0 / 1023.0 ) - 52.5;
    }

    // pos.y
    {
        boost::int32_t iy = val & MASK_9;
        pos->y = iy * ( 68.0 / 511.0 ) - 34.0;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
boost::int32_t
AudioCodec::posVelToBit31( const Vector2D & pos,
                           const Vector2D & vel ) const
{
    const double max_speed = ServerParam::i().ballSpeedMax();

    boost::int32_t rval = posToBit19( pos );

    rval <<= 6; // 6 bits shift for next info

    // vel.x value -> 6 bits (=[0,63])
    {
        //double vx = min_max( -2.7, vel.x, 2.7 );
        //vx += 2.7;
        //vx *= ( 63.0 / 5.4 ); // vx /= (5.4/63.0);
        double vx = min_max( -max_speed, vel.x, max_speed );
        vx += max_speed;
        vx *= ( 63.0 / ( max_speed*2.0 ) );
        rval |= static_cast< boost::int32_t >( rint( vx ) );
    }

    rval <<= 6; // 6 bits shift for next info

    // vel.y value -> 6 bits (=[0,63])
    {
        //double vy = min_max( -2.7, vel.y, 2.7 );
        //vy += 2.7;
        //vy *= ( 63.0 / 5.4 ); // vy /= (5.4/63.0);
        double vy = min_max( -max_speed, vel.y, max_speed );
        vy += max_speed;
        vy *= ( 63.0 / ( max_speed*2.0 ) );
        rval |= static_cast< boost::int32_t >( rint( vy ) );
    }

    return rval;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
AudioCodec::bit31ToPosVel( const boost::int32_t & val,
                           Vector2D * pos,
                           Vector2D * vel ) const
{
    const double max_speed = ServerParam::i().ballSpeedMax();

    bit19ToPos( val >> 12, pos );

    // vel.x
    {
        boost::int32_t ivx = ( val >> 6 ) & MASK_6;
        //vel->x = ivx * ( 5.4 / 63.0 ) - 2.7;
        vel->x = ivx * ( ( max_speed*2.0 ) / 63.0 ) - max_speed;
    }

    // vel.x
    {
        boost::int32_t ivy = val & MASK_6;
        //vel->y = ivy * ( 5.4 / 63.0 ) - 2.7;
        vel->y = ivy * ( ( max_speed*2.0 ) / 63.0 ) - max_speed;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::encodePosVelToStr5( const Vector2D & pos,
                                const Vector2D & vel,
                                std::string & to ) const
{
    boost::int64_t ival = posVelToBit31( pos, vel );

    //std::cout << " ival dec = "<< std::dec << ival << std::endl;
    //std::cout << " ival hex = "<< std::hex << ival << std::endl;

    return encodeInt64ToStr( ival, 5, to );
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::decodeStr5ToPosVel( const std::string & from,
                                Vector2D * pos,
                                Vector2D * vel ) const
{
    if ( from.length() != 5 )
    {
        return false;
    }

    boost::int64_t read_val = 0;

    if ( ! decodeStrToInt64( from, &read_val ) )
    {
        return false;
    }

    bit31ToPosVel( static_cast< boost::int32_t >( read_val ),
                   pos, vel );

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::encodePosToStr3( const Vector2D & pos,
                             std::string & to ) const
{
    boost::int32_t ival = posToBit18( pos );

    return encodeInt64ToStr( static_cast< boost::int64_t >( ival ),
                             3, to );
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::decodeStr3ToPos( const std::string & from,
                             Vector2D * pos ) const
{
    if ( from.length() != 3 )
    {
        return false;
    }

    boost::int64_t read_val = 0;

    if ( ! decodeStrToInt64( from, &read_val ) )
    {
        return false;
    }

    if ( pos )
    {
        bit18ToPos( static_cast< boost::int32_t >( read_val ), pos );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::encodeUnumPosToStr4( const int unum,
                                 const Vector2D & pos,
                                 std::string & to ) const
{
    if ( unum < 1 || 11 < unum )
    {
        return false;
    }

    boost::int64_t ival = posToBit19( pos );

    ival <<= 4;
    ival |= static_cast< boost::int64_t >( unum ); // 4 bits

    return encodeInt64ToStr( ival, 4, to );
}

/*-------------------------------------------------------------------*/
/*!

*/
bool AudioCodec::encodeUnumPosBodyToStr4( const int unum,
							  const Vector2D & pos,
							  const AngleDeg & bodyDir,
							  std::string & to ) const
{
	if ( unum < 1 || 22 < unum )
	{
		std::cerr << __FILE__ << ":" << __LINE__
				  << " ***ERROR*** encodeUnumPosBodyToStr4. illegal unum = "
				  << unum
				  << std::endl;
		dlog.addText( Logger::SENSOR,
				  "encodeUnumPosBodyToStr4. illegal unum = %d",
						unum );
		return false;
	}

	boost::int64_t ival = 0;
	double dval = 0.0;

	ival += unum - 1;

	dval = min_max( -52.5, pos.x, 52.5 ) + 52.5;
	ival *= 263;
	ival += static_cast< boost::int64_t >( rint( dval / 0.4 ) );

	dval = min_max( -34.0, pos.y, 34.0 ) + 34.0;
	ival *= 170;
	ival += static_cast< boost::int64_t >( rint( dval / 0.4 ) );

	dval = bodyDir.degree() + 180.0;
	ival *= 30; // = 360/12
	ival += static_cast< boost::int64_t >( rint( dval / 12.0 ) );

	std::string msg;
	msg.reserve( 4 );

	if ( ! AudioCodec::i().encodeInt64ToStr( ival, 4 , msg )
			|| msg.length() != 4 )
	{
		std::cerr << __FILE__ << ":" << __LINE__
				  << " ***ERROR*** encodeUnumPosBodyToStr4. "
				  << std::endl;

		dlog.addText( Logger::SENSOR,
					  "OnePlayerMessage. error!"
					  " unum=%d pos=(%.1f %.1f) body=%.1f",
					  unum,
					  pos.x, pos.y,
					  bodyDir.degree() );
		return false;
	}

	dlog.addText( Logger::SENSOR,
				  "encodeUnumPosBodyToStr4. success!. unum = %d pos=(%.1f %.1f)"
				  " body=%.1f -> [%s]",
				  unum,
				  pos.x, pos.y,
 				  bodyDir.degree(),
				  msg.c_str() );

	to += msg;

	return true;
}


/*-------------------------------------------------------------------*/
/*!

*/
bool AudioCodec::decodeStr4ToUnumPosBody( const std::string & from,
							  int * unum,
							  Vector2D * pos,
							  AngleDeg * bodyDir ) const
{
	boost::int64_t ival = 0;
	if ( from.size() != 4 )
	{
		std::cerr << "Boro baba 4 nist!!" << std::endl;
		return false;
	}

	if ( ! AudioCodec::i().decodeStrToInt64( from,
                                             &ival ) )
	{
		std::cerr << "decodeStr4ToUnumPosBody---->errorInDecoding" << std::endl;
		dlog.addText ( Logger::SENSOR , "decodeStr4ToUnumPosBody---->errorInDecoding" );
		return false;
    }

    int player_unum = Unum_Unknown;
    Vector2D player_pos;
    AngleDeg player_body;

	// 30=360/12
	player_body = static_cast< double >( ( ival % 30 ) * 12 - 180 );
	ival /= 30;

	// 170=68/0.4
	player_pos.y = ( ival % 170 ) * 0.4 - 34.0;
	ival /= 170;

	// 263>105/0.4=262.5
	player_pos.x = ( ival % 263 ) * 0.4 - 52.5;
	ival /= 263;

	player_unum = ( ival % 21 ) + 1;
	// ival /= 21

	*unum = player_unum;
	*pos = player_pos;
	*bodyDir = player_body;
	return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
AudioCodec::decodeStr4ToUnumPos( const std::string & from,
                                 int * unum,
                                 Vector2D * pos ) const
{
    if ( from.length() != 4 )
    {
        return false;
    }

    boost::int64_t read_val = 0;

    if ( ! decodeStrToInt64( from, &read_val ) )
    {
        return false;
    }

    boost::int32_t read_val32 = static_cast< boost::int32_t >( read_val );

    if ( unum )
    {
        *unum = static_cast< int >( read_val32 & MASK_4 );
    }

    if ( pos )
    {
        bit19ToPos( read_val32 >> 4, pos );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
std::string
AudioCodec::encodeCoordToStr2( const double & xy,
                               const double & norm_factor ) const
{
    // normalize value
    double tmp = min_max( -norm_factor, xy , norm_factor );
    tmp += norm_factor;

    // round
    tmp /= COORD_STEP_L2;

    int ival = static_cast< int >( rint( tmp ) );

    int i1 = ival % CHAR_SIZE;
    ival /= CHAR_SIZE;
    int i2 = ival % CHAR_SIZE;
    //std::cout << " posx -> " << ix1 << " " << ix2 << std::endl;

    if ( ival >= CHAR_SIZE )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::encodeCoordToStr2."
                  << " range over. value = " << xy
                  << " norm_factor = " << norm_factor
                  << std::endl;
        return std::string();
    }

    std::string msg;
    msg.reserve( 2 );

    try
    {
        msg += intToCharMap().at( i1 );
        msg += intToCharMap().at( i2 );
    }
    catch ( ... )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::encodeCoordToStr2."
                  << " Exception caught! "
                  << std::endl;
        return std::string();
    }

    return msg;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
AudioCodec::decodeStr2ToCoord( const char ch1,
                               const char ch2,
                               const double & norm_factor ) const
{
    const CharToIntCont::const_iterator end = charToIntMap().end();

    CharToIntCont::const_iterator it = charToIntMap().find( ch1 );
    if ( it == end )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::decodeStr2ToCoord()."
                  << " Unsupported character [" << ch1 << "]"
                  << std::endl;
        return ERROR_VALUE;
    }
    int i1 = it->second;

    it = charToIntMap().find( ch2 );
    if ( it == end )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::decodeStr2ToCoord()."
                  << " Unsupported character [" << ch2 << "]"
                  << std::endl;
        return ERROR_VALUE;
    }
    int i2 = it->second;

    return
        (
         static_cast< double >( i1 + i2 * CHAR_SIZE ) * COORD_STEP_L2
         - norm_factor
         );
}

/*-------------------------------------------------------------------*/
/*!

 */
std::string
AudioCodec::encodePosToStr4( const Vector2D & pos ) const
{
    std::string pos_str;
    pos_str.reserve( 4 );

    pos_str += encodeCoordToStr2( pos.x, X_NORM_FACTOR );
    pos_str += encodeCoordToStr2( pos.y, Y_NORM_FACTOR );

    if ( pos_str.length() != 4 )
    {
        std::cerr << "AudioCodec::encodePosToStr4(). "
                  << "Failed to encode " << pos
                  << std::endl;
        pos_str.clear();
    }

    return pos_str;
}

/*-------------------------------------------------------------------*/
/*!

 */
Vector2D
AudioCodec::decodeStr4ToPos( const std::string & msg ) const
{
    if ( msg.length() < 4 )
    {
        std::cerr << "AudioCodec::decodeStr4ToPos(). "
                  << "message length is too short " << msg
                  << std::endl;
        return Vector2D::INVALIDATED;
    }

    Vector2D pos;

    pos.x = decodeStr2ToCoord( msg[0], msg[1], X_NORM_FACTOR );
    if ( pos.x == ERROR_VALUE )
    {
        std::cerr << "AudioCodec::decodeStr4ToPos(). "
                  << "Unexpected x value " << msg
                  << std::endl;
        return Vector2D::INVALIDATED;
    }

    pos.y = decodeStr2ToCoord( msg[2], msg[3], Y_NORM_FACTOR );
    if ( pos.y == ERROR_VALUE )
    {
        std::cerr << "AudioCodec::decodeStr4ToPos(). "
                  << "Unexpected x value " << msg
                  << std::endl;
        return Vector2D::INVALIDATED;
    }

    return pos;
}

/*-------------------------------------------------------------------*/
/*!

*/
char
AudioCodec::encodeSpeedToChar( const double & val ) const
{
    // normalize value
    double tmp = min_max( -SPEED_NORM_FACTOR, val , SPEED_NORM_FACTOR );
    tmp += SPEED_NORM_FACTOR;

    // round
    tmp /= SPEED_STEP_L1;

    int ival = static_cast< int >( rint( tmp ) );

    try
    {
        return  intToCharMap().at( ival );
    }
    catch ( ... )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::encodeSpeedL1."
                  << " Exception caught! Failed to encode"
                  << std::endl;
        return '\0';
    }

    return '\0';
}

/*-------------------------------------------------------------------*/
/*!

*/
double
AudioCodec::decodeCharToSpeed( const char ch ) const
{
    CharToIntCont::const_iterator it = charToIntMap().find( ch );
    if ( it == charToIntMap().end() )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** AudioCodec::decodeSpeedL1."
                  << " Unsupported character [" << ch << "]"
                  << std::endl;
        return ERROR_VALUE;
    }

    return
        ( static_cast< double >( it->second ) * SPEED_STEP_L1
          - SPEED_NORM_FACTOR
          );
}

long int AudioCodec::encodeUnumPosToLongInt( int unum , Vector2D & pos)
{
	WindowCoding windowCoding;
	float xPrec = (float)windowCoding.windowLength / (float)(windowCoding.windowXStates - 1);
	float yPrec = (float)windowCoding.windowWidth / (float)(windowCoding.windowYStates - 1);
	float tempX = pos.x / xPrec;
	float tempY = pos.y / yPrec;
	int xQuant = (int)(tempX + .5);
	int yQuant = (int)(tempY + .5);
	int xyPlate = (windowCoding.windowXStates * windowCoding.windowYStates);
	long int sum = xyPlate * (unum - 1);
	sum += (yQuant) * windowCoding.windowXStates;
	sum += xQuant;
	return sum;
}

bool AudioCodec::decodeLongIntToUnumPos( long int middleCode, Vector2D & pos , int &unum)
{
	WindowCoding windowCoding;
	int xyPlate = ( windowCoding.windowXStates * windowCoding.windowYStates );
	unum = middleCode / xyPlate + 1;
	float xPrec = (float)windowCoding.windowLength / (float)(windowCoding.windowXStates - 1);
	float yPrec = (float)windowCoding.windowWidth / (float)(windowCoding.windowYStates - 1);
	int rest = middleCode % xyPlate;
	int xQuant = rest % windowCoding.windowXStates;
	int yQuant = rest / windowCoding.windowXStates;
	pos.x = xQuant * xPrec;
	pos.y = yQuant * yPrec;
	return true;
}

long long int AudioCodec::mergeFour( long long int first, long long int second, long long int third, long long int fourth)
{
	WindowCoding windowCoding;
	long long int base = (long long int)windowCoding.windowXStates * (long long int)windowCoding.windowYStates * (long long int)windowCoding.unumStates;
	dlog.addText ( Logger::PASS , "base\t%d" , base );
	dlog.addText ( Logger::PASS , "first\t%d" , first );
	dlog.addText ( Logger::PASS , "second\t%d" , base * second );
	dlog.addText ( Logger::PASS , "third\t%d" , base * base * third );
	dlog.addText ( Logger::PASS , "fourth\t%d" , base * base * base * fourth );
	long long int answer = base * base * base * fourth +
							base * base * third +
							base * second +
							first;
	return answer;
}

void AudioCodec::splitInFour(long long int bigCode, long int &firstAns, long int &secondAns, long int &thirdAns, long int &fourthAns)
{
	WindowCoding windowCoding;
	long int base = (long int)windowCoding.windowXStates * (long int)windowCoding.windowYStates * (long int)windowCoding.unumStates;
	firstAns = bigCode % base;
	bigCode = bigCode / base;
	secondAns = bigCode % base;
	bigCode /= base;
	thirdAns = bigCode % base;
	bigCode /= base;
	fourthAns = bigCode; 
}

std::string AudioCodec::putLongLongInString( long long int code )
{
	int baseCount[9];
	int base = CHAR_SET.length();
	for (int i = 0; i < 9; i++)
	{
		long long int temp = 1;
		for (int j = 0; j < 9 - i - 1; j++)
			temp *= base;
		baseCount[i] = code / temp;
		code = code % temp;
	}
	std::string returnValue = "";
	dlog.addText ( Logger::PASS , "code\t%lld" , code );
	for ( int i = 0; i < 9; i++)
	{
		dlog.addText ( Logger::PASS , "baseCount[%d]\t%d", i , baseCount[i] );
		returnValue += CHAR_SET[baseCount[i]];
	}
	return returnValue;
}

long long int AudioCodec::convertStringToLongLongInt( std::string codedString )
{
	long long int answer = 0;
	for ( unsigned int i = 0; i < codedString.length(); i++ )
	{
		int thisCharPos = 0;
		for ( unsigned int j = 0; j < CHAR_SET.length(); j++ )
			if ( CHAR_SET[j] == codedString[i] )
			{
				thisCharPos = j;
				break;
			}
		answer += thisCharPos * pow( CHAR_SET.length() , 8 - i );
	}
	return answer;
}

Vector2D AudioCodec::changeCoordinates( rcsc::Vector2D realPosition )
{
	WindowCoding windowCoding;
	Vector2D fakePosition = realPosition;
	fakePosition.x += windowCoding.fieldLength / 2.f;
	fakePosition.y += windowCoding.fieldWidth / 2.f;
	float startX = (float)((int)( fakePosition.x / windowCoding.windowLength )) * windowCoding.windowLength;
	float startY = (float)((int)( fakePosition.y / windowCoding.windowWidth )) * windowCoding.windowWidth;
	Vector2D answerPosition;
	answerPosition.x = fakePosition.x - startX;
	answerPosition.y = fakePosition.y - startY;
	return answerPosition;
}

int AudioCodec::getBestPositionInWindows( Vector2D & answerPosition , Vector2D receivedPosition , Vector2D lastPosition , int posCount )
{
	WindowCoding windowCoding;
	Vector2D best = Vector2D( 0.f , 0.f );
	float minDist = 9999.f;
	float posCountCircle = posCount * 0.95;
	int goodCount = 0;
	for ( int i = 0; i < windowCoding.windowsXCount; i++ )
	{
		for ( int j = 0; j < windowCoding.windowsYCount; j++ )
		{
			Vector2D thisPos( - windowCoding.fieldLength / 2.f , - windowCoding.fieldWidth / 2.f );
			thisPos.x += i * windowCoding.windowLength;
			thisPos.y += j * windowCoding.windowWidth;
			thisPos = thisPos + receivedPosition;
			float dist = thisPos.dist( lastPosition );
			if ( dist > posCountCircle )
				continue;
			goodCount ++;
			if ( dist < minDist )
			{
				minDist = dist;
				best = thisPos;
			}
		}
	}
	answerPosition = best;
	return goodCount;
}

/*-------------------------------------------------------------------*/
/*!

*/
char
AudioCodec::unum2hex( const int unum )
{
    switch ( unum ) {
	case 0: return '0'; break;
    case 1: return '1'; break;
    case 2: return '2'; break;
    case 3: return '3'; break;
    case 4: return '4'; break;
    case 5: return '5'; break;
    case 6: return '6'; break;
    case 7: return '7'; break;
    case 8: return '8'; break;
    case 9: return '9'; break;
    case 10: return 'A'; break;
    case 11: return 'B'; break;
    default:
        break;
    }

    return '\0';
}

/*-------------------------------------------------------------------*/
/*!

*/
int
AudioCodec::hex2unum( const char hex )
{
    switch ( hex ) {
	case '0': return 0; break;
    case '1': return 1; break;
    case '2': return 2; break;
    case '3': return 3; break;
    case '4': return 4; break;
    case '5': return 5; break;
    case '6': return 6; break;
    case '7': return 7; break;
    case '8': return 8; break;
    case '9': return 9; break;
    case 'A': return 10; break;
    case 'B': return 11; break;
    default:
        break;
    }

    return Unum_Unknown;
}

std::string
AudioCodec::cycle2str2 ( const int cycle )
{
	std::string result = "00";
	result[0] = ((char)( cycle % 100 ) / 10 ) + '0';
	result[1] = ((char)cycle % 10) + '0';
	return result;
}

std::string 
AudioCodec::cycle2str ( const int cycle )
{
	std::string result = "000";
	int importantPart = cycle % 1000;
	result[0] = ((char)( importantPart / 100 ))+'0';
	result[1] = ((char)( ( importantPart / 10 ) % 10 ))+'0';
	result[2] = ((char)( importantPart % 10 ))+'0';
	return result;
}

std::string 
AudioCodec::cycle2str4 ( const int cycle )
{
	std::string result = "0000";
	result[0] = ((char)( cycle / 1000 ))+'0';
	result[1] = ((char)( ( cycle / 100 )% 10 ))+'0';
	result[2] = ((char)( ( cycle / 10 ) % 10 ))+'0';
	result[3] = ((char)( cycle % 10 )) + '0';
	return result;
}

}
